/*Si implementi in linguaggio C in ambiente Unix/Linux un programma che simuli il comportamento
di una stazione ferroviaria mediante processi concorrenti, secondo le seguenti specifiche.
Una stazione ferroviaria è collegata con la rete ferroviaria da:
 - NLE (Numero Linee di Entrata) coppie di binari di entrata (LE, Linee di Entrata)
 - NLU (Numero Linee di Uscita) coppie di binari in uscita (LU, Linee di Uscita)
ma è attraversata da un’unica coppia di binari (e quindi è attraversabile da un solo treno alla volta).
Ogni treno che giunge alla stazione da una linea ferroviaria di entrata, può entrare nella stazione
se non ci sono altri treni nella stazione stessa. Inoltre può lasciare la stazione solo se la linea
di uscita selezionata è libera.
Si osservi che ogni treno arriva alla stazione in un istante di tempo casuale, si ferma in stazione per
un tempo casuale e decide di proseguire su una linea scelta anch’essa a caso, occupandola infine per un
numero casuale di secondi.
Quando la stazione viene liberata da un treno e più treni sono in attesa di entrare viene fatto procedere
quello che arriva dalla linea di entrata con indice più basso.

Suggerimenti:
 - si realizzi un processo per la stazione e uno per ciascun treno
 - si realizzi inoltre una pipe per le linee ferroviarie in entrata alla stazione e una per le linee di uscita
 - l’ingresso in stazione di ciascun treno sia inoltre controllato da un semaforo. I semafori siano realizzati attraverso pipe.*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <string.h>

#define NLE 5
#define NLU 3

/* File che contiene il numero di treni in attesa in ogni linea di entrata e relativo semaforo
   Memorizzo tali informazioni in un file e non in una variabile globale, un vettore, poiche'
   ogni processo modificherebbe la propria copia del vettore in quanto i processi hanno spazio di
   indirizzamento diverso, ed il processo stazione non si accorgerebbe di nulla */
char file[] = "n_treni.txt";
int sem_n_treni[2];

int linee_entrata[NLE][2]; //semafori per ogni linea d'entrata
int linee_uscita[NLU][2]; //semafori per ogni linea d'uscita
int sem_stazione[2];	//semaforo per l'accesso alla stazione
int sem_generale[2];         //ogni treno fa una post su questo semaforo

typedef void SigFunc (int);
void set_handler (int, SigFunc);
void handler(int);
void sem_wait(int *);
void sem_post(int *);
void treno();
void stazione();
void dec(int);
void inc(int);
int  get(int);

int main()
{
    int i;    
    
    srand(time(0)+getpid());
    FILE * f = fopen(file, "w");
    for(i = 0; i < NLE; i++)
	fprintf(f, "%d=0\n", i);    
    fflush(f);
    fclose(f);
    
    //creo il semaforo-pipe del numero di treni
    if(pipe(sem_n_treni)){
	printf("Errore nella creazione della pipe del numero dei treni!\n");
	exit(EXIT_FAILURE);
    }
    sem_post(sem_n_treni);
    
    //creo il semaforo-pipe della stazione
    if(pipe(sem_stazione)){
	printf("Errore nella creazione della pipe della stazione!\n");
	exit(EXIT_FAILURE);
    }
    sem_post(sem_stazione);
    
    //creo il semaforo generale
    if(pipe(sem_generale)){
	printf("Errore nella creazione della pipe generale!\n");
	exit(EXIT_FAILURE);
    }
    
    //creo le pipe in entrata
    for(i = 0; i < NLE; i++){
	if(pipe(linee_entrata[i])){
	    printf("Errore nella creazione della pipe in entrata %d!\n", i);
	    exit(EXIT_FAILURE);
	}
    }
    
    //creo le pipe in uscita
    for(i = 0; i < NLU; i++){
	if(pipe(linee_uscita[i])){
	    printf("Errore nella creazione della pipe in uscita %d!\n", i);
	    exit(EXIT_FAILURE);
	}
    }
    
    //Libero tutte le linee d' uscita per poter iniziare
    for(i = 0; i < NLU; i++)
	sem_post(linee_uscita[i]);
    
    set_handler(SIGINT, handler); /* Gestisce Ctrl-C */
    set_handler(SIGTSTP, handler);/* Gestisce Ctrl-Z */
    
    //Il processo main crea il processo stazione e poi crea i processi del treno
    if(fork() == 0)
    {
	//Sono il processo della stazione
	stazione();
    }
    else
    {
	//sono il processo main, creo i vari processi treni
	while(1)
	{
	    if(fork() == 0)
	    {
		treno();		
		/* finisco il ciclo, altrimenti il processo treno che ha appena finito
		   continua a cicliare nel while(1) e a creare a sua volta altri treni*/
		break;
	    }
	    else
		sleep(rand()%5+1);
	}
    }
    return EXIT_SUCCESS;
}

void sem_wait(int * pipe)
{
    char buffer[2];
    if (read(pipe[0], &buffer, 1) != 1) {
	printf ("Error on sem_wait\n");
	exit (EXIT_FAILURE);
    }
}

void sem_post(int * pipe)
{    
    if (write(pipe[1], "X", 1) != 1) {
	printf ("Error on sem_post\n");
	exit (EXIT_FAILURE);
    }
}

void treno ()
{
    srand(time(0)+getpid());
    
    //Scegliamo una linea di entrata e d'uscita
    int in = rand()%NLE;
    int out = rand()%NLU;
    
    int me = getpid()-getppid()-1;
    
    //segnalo che sono arrivato ed incremento il contatore
    //del numero di treni nella mia linea di entrata
    sem_wait(sem_n_treni);    
    inc(in);
    sem_post(sem_n_treni);    
    
    sem_post(sem_generale);
    
    printf("Treno %d aspetta al binario %d per entrare in stazione.\n", me, in);
    sem_wait(linee_entrata[in]);
    
    printf("Treno %d in stazione.\n", me);
    
    //sono in stazione. ci resto un po'
    sleep(rand()%5+1);
    
    printf("Treno %d attende al binario d'uscita %d.\n", me, out);
    //aspetto che la mia linea di uscita si liberi
    sem_wait(linee_uscita[out]);
    
    printf("Treno %d in uscita.\n", me);
    //libero la stazione
    sem_post(sem_stazione);
    
    //resto per un po' sulla linea d'uscita
    sleep(rand()%5+1);
    
    printf("Treno %d uscito.\n", me);
    sem_post(linee_uscita[out]);
}

void stazione ()
{
    int i, binario = -1;
    while(1)
    {
	//Vediamo se c'e' almeno un treno in attesa
	sem_wait(sem_generale);

	//Almeno un treno c'e', facciamone passare uno
	//corrispondente al binario con indice minore	
	sem_wait(sem_n_treni);
	for(i = 0; i < NLE; i++)
	    if(get(i) > 0)
		binario = i;
	dec(binario);
	sem_post(sem_n_treni);
	
	//libero un treno di quella linea
	sem_post(linee_entrata[binario]);

	//e aspetto che libera la stazione
	sem_wait(sem_stazione);	
    }
}

void dec (int index)
{
    char buffer[20];
    FILE * f = fopen(file, "r+");
    int num, i, val = 0;
    
    for(i = 0; i < index+1; i++)
	fgets(buffer, 20, f);
    
    for(i = 0; buffer[i] != '='; i++);
    i++;
    for(; buffer[i] != '\n'; i++)
	buffer[val++] = buffer[i];
    buffer[val] = '\0';
    
    num = atoi(buffer)-1;
    
    rewind(f);
    
    //fittizio
    for(i = 0; i < index; i++)
	fgets(buffer, 20, f);
    
    fprintf(f, "%d=%d\n", index, num);
    
    fflush(f);
    fclose(f);
}

void inc (int index)
{
    char buffer[20];
    FILE * f = fopen(file, "r+");
    int num, i, val = 0;
    for(i = 0; i < index+1; i++)
	fgets(buffer, 20, f);
    for(i = 0; buffer[i] != '='; i++);
    i++;
    for(; buffer[i] != '\n'; i++)
	buffer[val++] = buffer[i];    
    buffer[val] = '\0';

    num = atoi(buffer)+1;    
    
    rewind(f);

    
    //fittizio
    for(i = 0; i < index; i++)
	fgets(buffer, 20, f);    
    fprintf(f, "%d=%d\n", index, num);

    fflush(f);
    fclose(f);
}

int get (int index)
{
    char buffer[20];
    FILE * f = fopen(file, "r+");
    int i, val = 0;
    for(i = 0; i < index+1; i++)
	fgets(buffer, 20, f);
    for(i = 0; buffer[i] != '='; i++);
    i++;
    for(; buffer[i] != '\n'; i++)
	buffer[val++] = buffer[i];    
    buffer[val] = '\0';
    
    fclose(f);
    return atoi(buffer);
}

void set_handler (int signo, SigFunc handler)
{
    /* Utilizziamo sa_handler, non sa_sigaction, pertanto
    non utilizziamo il flag SA_SIGINFO.
    Il segnale passato come argomento viene mascherato
    di default, a meno che non si utilizzi il flag
    SA_NODEFER o SA_NOMASK in sa_flags */
    
    struct sigaction act;
    memset(&act, 0, sizeof(act));
    act.sa_handler = handler;
    //act.sa_flags = SA_NODEFER; non ha effetto
    if(sigaction(signo, &act, NULL) == -1)
	printf("ERROR on sigaction\n");
}

void handler(int n)
{
    char filename[1024];
    sprintf(filename, "./%s", file);
    remove(filename);
    exit(EXIT_SUCCESS);
}
