/*Due direttori si definiscono equivalenti se contengono le stesse informazioni,
ovvero se, ricorsivamente, hanno lo stesso contenuto.
Si realizzi, in linguaggio C, in ambiente Unix/Linux, un programma concorrente
in grado di determinare l’equivalenza di due direttori.
Il programma deve:
  - ricevere sulla riga di comando il pathname completo (path+nome) di due direttori
  - duplicarsi in due processi, in modo che il padre visiti ricorsivamente uno dei
    due direttorio e il figlio visiti ricorsivamente l’altro direttorio
  - i due processi (padre e figlio) devono sincronizzarsi a ogni file/direttorio
    rintracciato nella struttura gerarchica visitata per verificare che esso sia lo stesso.
    In caso affermativo continuano il procedimento.
    In caso negativo segnalano l’incongruenza e terminano.

Si limiti il confronto di equivalenza ai soli nomi dei file/direttori
(non ci si occupi della dimensione, dei diritti di accesso, etc.).

Suggerimento
Per ogni direttorio/file rintracciato nella struttura, uno dei due processi trasferisce su
una pipe all’altro processo il nome del direttorio/file trovato e quindi aspetta su un semaforo
che l’altro processo verifichi le informazioni trasferite e lo svegli. */

/* Si suppone che la readdir() proceda allo stesso modo in entrambi i processi */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <signal.h>

#define MAX_LEN 1024
#define OK   1
#define FAIL 0

int status;
int fd[2];
int sem[2];

typedef void SigFunc (int);
void set_handler (int, SigFunc);
void handlerOK(int);
void handlerFAIL(int);
void processo1(const char *);
void processo2(const char *, pid_t);
void sem_wait(int *);
void sem_post(int *);

int main(int argc, char ** argv)
{
    pid_t figlio1, figlio2;
    int exit_value;
    if(argc != 3){
	printf("Usage: %s <dir_1> <dir_2>\n", argv[0]);
	exit(EXIT_FAILURE);
    }
    
    if(pipe(fd)){
	printf("Errore nella creazione della pipe!\n");
	exit(EXIT_FAILURE);
    }
    if(pipe(sem)){
	printf("Errore nella creazione del semaforo!\n");
	exit(EXIT_FAILURE);
    }
    
    if((figlio1 = fork()) == 0)
    {
	set_handler(SIGUSR1, handlerOK);
	set_handler(SIGUSR2, handlerFAIL);
	//Questo processo scrive nella pipe
	close(fd[0]);
	processo1(argv[1]);
    }
    else
    {
	if((figlio2 = fork()) == 0){
	    //Questo processo legge dalla pipe
	    close(fd[1]);
	    processo2(argv[2], figlio1);
	}else{
	    close(fd[0]);
	    close(fd[1]);
	    wait(NULL);
	    waitpid(-1, &exit_value, 0);
	}
    }
    if(exit_value == EXIT_SUCCESS)
	printf("Le directory sono equivalenti.\n");
    else
	printf("Le directory non sono equivalenti.\n");
    
    return EXIT_SUCCESS;
}

void handlerOK(int n)
{
    status = OK;
}

void handlerFAIL(int n)
{
    status = FAIL;
}

void processo1(const char * path)
{
    DIR * dir;
    struct dirent * dir_struct;
    char buffer[MAX_LEN];
    
    dir = opendir(path);
    if(dir == NULL)
    {
	sprintf(buffer, "Error on opening '%s'", path);
	perror(buffer);
	exit(EXIT_FAILURE);
    }
    
    while((dir_struct = readdir(dir)) != NULL)
    {
	if(!strcmp(dir_struct->d_name, ".") || !strcmp(dir_struct->d_name, ".."))
	    continue;
	
	if(dir_struct->d_type == DT_DIR || dir_struct->d_type == DT_REG)
	{   
	    sprintf(buffer, "%s", dir_struct->d_name);
	    write(fd[1], buffer, strlen(buffer)+1);
	    pause();    
	    if(status == FAIL) exit(EXIT_FAILURE);	    
	    sem_post(sem);
	}
	if(dir_struct->d_type == DT_DIR){
	    sprintf(buffer, "%s/%s", path, dir_struct->d_name);
	    processo1(buffer);
	}
    }
    exit(EXIT_SUCCESS);
}

void processo2(const char * path, pid_t fratello)
{
    DIR * dir;
    struct dirent * dir_struct;
    char buffer[MAX_LEN];
    
    dir = opendir(path);
    if(dir == NULL)
    {
	sprintf(buffer, "Error on opening '%s'", path);
	perror(buffer);
	exit(EXIT_FAILURE);
    }
    
    while((dir_struct = readdir(dir)) != NULL)
    {
	if(!strcmp(dir_struct->d_name, ".") || !strcmp(dir_struct->d_name, ".."))
	    continue;
	
	if(dir_struct->d_type == DT_DIR || dir_struct->d_type == DT_REG)
	{   
	    read(fd[0], buffer, MAX_LEN);
	    if(!strcmp(buffer, dir_struct->d_name))
		kill(fratello, SIGUSR1);
	    else{
		kill(fratello, SIGUSR2);
		exit(EXIT_FAILURE);
	    }
	    sem_wait(sem);
	}
	if(dir_struct->d_type == DT_DIR){
	    sprintf(buffer, "%s/%s", path, dir_struct->d_name);
	    processo2(buffer, fratello);
	}
    }
    exit(EXIT_SUCCESS);
}

void set_handler (int signo, SigFunc handler)
{
    /* Utilizziamo sa_handler, non sa_sigaction, pertanto
    non utilizziamo il flag SA_SIGINFO.
    Il segnale passato come argomento viene mascherato
    di default, a meno che non si utilizzi il flag
    SA_NODEFER o SA_NOMASK in sa_flags */
    
    struct sigaction act;
    memset(&act, 0, sizeof(act));
    act.sa_handler = handler;
    //act.sa_flags = SA_NODEFER; non ha effetto
    if(sigaction(signo, &act, NULL) == -1)
	printf("ERROR on sigaction\n");
}

void sem_wait(int * pipe)
{
    char buffer[2];
    if (read(pipe[0], &buffer, 1) != 1) {
	printf ("Error on sem_wait\n");
	exit (EXIT_FAILURE);
    }
}

void sem_post(int * pipe)
{    
    if (write(pipe[1], "X", 1) != 1) {
	printf ("Error on sem_post\n");
	exit (EXIT_FAILURE);
    }
}
